from channels.middleware import BaseMiddleware
from channels.db import database_sync_to_async
from rest_framework_simplejwt.tokens import AccessToken, TokenError
from django.contrib.auth.models import AnonymousUser
from django.contrib.auth import get_user_model

User = get_user_model()


@database_sync_to_async
def get_user_from_jwt(token):
    try:
        access_token = AccessToken(token)
        user_id = access_token["user_id"]
        return User.objects.get(id=user_id)
    except (TokenError, User.DoesNotExist) as e:
        print(f"[JWT ERROR] {e}")
        return None


class TokenAuthMiddleware(BaseMiddleware):
    async def __call__(self, scope, receive, send):
        headers = dict(scope.get("headers", {}))
        token = None

        if b"authorization" in headers:
            auth_header = headers[b"authorization"].decode()
            if auth_header.lower().startswith("bearer "):
                token = auth_header.split(" ")[1]

        user = await get_user_from_jwt(token) if token else None

        if user is None:
            print("[AUTH] WebSocket ulanish rad etildi: foydalanuvchi aniqlanmadi yoki token noto‘g‘ri.")
            await send({
                "type": "websocket.close",
                "code": 4001
            })
            return

        scope["user"] = user
        return await super().__call__(scope, receive, send)



# test_project/asgi.py

import os
import django
from channels.routing import ProtocolTypeRouter, URLRouter
from channels.auth import AuthMiddlewareStack
from django.core.asgi import get_asgi_application

os.environ.setdefault("DJANGO_SETTINGS_MODULE", "test_project.settings")
django.setup()

import test_project.routing  # routing.py import qilish

application = ProtocolTypeRouter({
    "http": get_asgi_application(),
    "websocket": AuthMiddlewareStack(
        URLRouter(
            test_project.routing.websocket_urlpatterns
        )
    ),
})



# djwsbridge/consumers.py
from channels.generic.websocket import AsyncWebsocketConsumer
from django.conf import settings
from django.contrib.auth import get_user_model
from asgiref.sync import sync_to_async
import json

from .utils import get_chat_model, get_message_model, filter_valid_fields
from django.core.exceptions import ImproperlyConfigured

User = get_user_model()
ALLOWED_MESSAGE_TYPES = getattr(settings, "DJWSBRIDGE_ALLOWED_TYPES", None)


class UserConsumer(AsyncWebsocketConsumer):
    async def connect(self):
        user = self.scope.get("user")
        if user is None or not user.is_authenticated:
            await self.close()
            return

        self.user = user
        self.user_id = user.id
        self.group_name = f"user_{self.user_id}"

        await self.channel_layer.group_add(
            self.group_name,
            self.channel_name
        )
        await self.accept()

    async def receive(self, text_data):
        try:
            data = json.loads(text_data)
        except json.JSONDecodeError:
            await self.send_json({"error": "Invalid JSON format"})
            return

        msg_type = data.get("type")
        # Message type tekshirish (agar ALLOWED_MESSAGE_TYPES mavjud bo‘lsa)
        if ALLOWED_MESSAGE_TYPES is not None and msg_type not in ALLOWED_MESSAGE_TYPES:
            await self.send_json({"error": f"Invalid message type: {msg_type}"})
            return

        # Chat va Message modellar mavjudligini tekshirish
        Chat = None
        Message = None
        try:
            Chat = get_chat_model()
        except ImproperlyConfigured:
            pass

        try:
            Message = get_message_model()
        except ImproperlyConfigured:
            pass

        recipient = None
        if "recipient" in data:
            try:
                recipient = int(data["recipient"])
                recipient = await User.objects.aget(id=recipient)
            except (ValueError, User.DoesNotExist):
                await self.send_json({"error": "Invalid or non-existent recipient_id"})
                return

            if getattr(recipient, "is_blocked", False):
                await self.send_json({"error": "Recipient is blocked"})
                return

        # Chat va message modeli mavjud bo‘lsa, xabarni saqlaymiz
        if Chat and Message and recipient:
            chat = await sync_to_async(self.get_or_create_chat)(Chat, self.user_id, recipient.id)

            # sender va recipient ni instance ko‘rinishida qo‘shamiz
            data_for_model = data.copy()
            data_for_model["sender"] = self.user
            data_for_model["recipient"] = recipient
            data_for_model["chat"] = chat

            try:
                valid_data = filter_valid_fields(Message, data_for_model)
                await sync_to_async(Message.objects.create)(**valid_data)
            except ValueError as e:
                await self.send_json({"error": str(e)})
                return

        # Har doim xabarni uzatamiz: avvalo boshqa foydalanuvchiga
        if recipient:
            await self.channel_layer.group_send(
                f"user_{recipient.id}",
                {
                    "type": "send_message",
                    "data": {
                        "sender": self.user_id,
                        **data
                    }
                }
            )

        # O‘ziga ham yuboriladi (feedback)
        await self.channel_layer.group_send(
            self.group_name,
            {
                "type": "send_message",
                "data": {
                    "sender": self.user_id,
                    **data
                }
            }
        )

    @staticmethod
    def get_or_create_chat(Chat, user1_id, user2_id):
        chat = Chat.objects.filter(participants__id=user1_id) \
            .filter(participants__id=user2_id) \
            .distinct().first()
        if not chat:
            chat = Chat.objects.create()
            chat.participants.add(user1_id, user2_id)
        return chat

    async def disconnect(self, close_code):
        await self.channel_layer.group_discard(
            self.group_name,
            self.channel_name
        )

    async def send_message(self, event):
        data = event.get("data", {})
        if not isinstance(data, dict):
            await self.send_json({"error": "Invalid message format"})
            return
        await self.send_json(data)

    async def send_json(self, content):
        await self.send(text_data=json.dumps(content))



# djwsbridge/utils.py
from django.apps import apps
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured


def get_chat_model():
    model_path = getattr(settings, "DJWSBRIDGE_CHAT_MODEL", None)
    if not model_path:
        raise ImproperlyConfigured("DJWSBRIDGE_CHAT_MODEL is not defined in settings.")
    return apps.get_model(model_path)


def get_message_model():
    model_path = getattr(settings, "DJWSBRIDGE_MESSAGE_MODEL", None)
    if not model_path:
        raise ImproperlyConfigured("DJWSBRIDGE_MESSAGE_MODEL is not defined in settings.")
    return apps.get_model(model_path)


def filter_valid_fields(model_class, data: dict) -> dict:
    field_names = {f.name for f in model_class._meta.get_fields()}
    filtered_data = {k: v for k, v in data.items() if k in field_names}
    invalid_keys = set(data.keys()) - field_names

    if invalid_keys:
        raise ValueError(f"Invalid fields for model {model_class.__name__}: {', '.join(invalid_keys)}")
    return filtered_data



from django.conf import settings
from djwsbridge.consumers import UserConsumer
from django.urls import re_path


def get_websocket_urlpatterns():
    try:
        prefix = settings.DJWSBRIDGE_WS_PREFIX
        path = settings.DJWSBRIDGE_WS_PATH
    except AttributeError:
        raise RuntimeError(
            "DJWSBRIDGE_WS_PREFIX topilmadi. Iltimos, settings.py faylga quyidagicha yozing:\n\n"
            "DJWSBRIDGE_WS_PREFIX = 'ws/'"
        )

    return [
        re_path(rf'^{prefix}/{path}/$', UserConsumer.as_asgi()),
    ]



from asgiref.sync import async_to_sync
from channels.layers import get_channel_layer

def send_ws_message(user_id: int, data: dict):
    channel_layer = get_channel_layer()
    if channel_layer is None:
        raise RuntimeError("Channel layer topilmadi. Django Channels sozlanmagan bo‘lishi mumkin.")
    group_name = f"user_{user_id}"
    async_to_sync(channel_layer.group_send)(
        group_name,
        {
            "type": "send_message",
            "data": data,
        }
    )



from setuptools import setup, find_packages

setup(
    name="djwsbridge",
    version="0.1.1",
    description="Real-time WebSocket bridge for Django (using Channels)",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Asadbek",
    author_email="email@asadbektuygunov9@gmail.com",
    url="https://github.com/asadbek000002/djwsbridge",  # bo‘sh qoldirsa ham bo‘ladi
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        "Django>=4.0",
        "channels>=4.0",
        "asgiref>=3.5",
        "channels_redis>=4.0",
        "daphne>=4.0"
    ],
    classifiers=[
        "Framework :: Django",
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.8",
)
